// T-Code Random Stroker v2.0
// TempestVR 31-10-22
// Please copy, share, learn, innovate, give attribution.
// This plugin animates the hips of a female with a series of fun randomised motions.
// It also then outputs a synced stream of data to the Serial Port that can be used 
// to drive a generic device (T-code)
//
// 31-10-20 v2.0 Six axis T-stroke version implemented

// Import libraries
using System.Linq;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System;
using System.Diagnostics;
using System.IO.Ports;
using System.Threading;
using System.Text;

// The magic starts here
namespace SerialCtrl {
    public class RandomStroker : MVRScript {
		
        // Name of plugin
        public static string pluginName = "Random Stroker";
		
		// Serial Variables
        protected JSONStorableStringChooser portChooser;
        private string[] portNames;
        private string portSelected = "None";
		private int baudRate;
        private int sendThisChannelAnyway = 0;
        protected UIDynamicButton startButton;
        protected UIDynamicButton stopButton;
        protected JSONStorableBool autoLoadToggle;
        protected UIDynamicButton loadButton;
        protected UIDynamicButton saveButton;
        protected JSONStorableString serialReport;
        protected UIDynamicTextField serialOutWindow;
		protected JSONStorableBool suppressRepeats;
        protected JSONStorableBool logSerialIn;

        // TCode axis variables
        private List<TCodeAxis> axesList;
		
		// TCode Variables
		protected JSONStorableFloat OutMaxX;
        protected JSONStorableFloat OutMinX;
        protected JSONStorableFloat OutMaxRY;
        protected JSONStorableFloat AdjustRY;
        protected JSONStorableFloat OutMaxRZ;
        protected JSONStorableFloat AdjustRZ;
		protected JSONStorableFloat OutMaxV0;
        protected JSONStorableBool InvertX;
        protected JSONStorableBool InvertRY;
        protected JSONStorableBool InvertRZ;
        protected JSONStorableBool OverrideBoolX;
        protected JSONStorableBool OverrideBoolRY;
        protected JSONStorableBool OverrideBoolRZ;
		protected JSONStorableBool OverrideBoolV0;
        protected JSONStorableFloat OverrideX;
        protected JSONStorableFloat OverrideRY;
        protected JSONStorableFloat OverrideRZ;
		protected JSONStorableFloat OverrideV0;
		
		// Stroker UI variables
		protected JSONStorableString strokerReport;
        protected UIDynamicTextField strokerWindow;
		protected JSONStorableFloat intensityLevel;
		protected JSONStorableFloat intensityEscalate;
		protected JSONStorableFloat weightTease;
		protected JSONStorableFloat weightGrind;
		protected JSONStorableFloat weightLong;
		protected JSONStorableFloat weightShort;
		protected JSONStorableFloat randomVariation;
		protected JSONStorableFloat adjustGirlUp;
		protected JSONStorableFloat adjustGirlForward;
		protected JSONStorableFloat scaleTranslation;
		protected JSONStorableFloat scaleRotation;
		
		// Random stroker movement modes
		// Short strokes 
		private float[][,] ShortStrokes = new float[6][,]
		{
			new float[,] { // low roll forward
				{ 0.0f, 0.5f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.4f, 0.6f, 1f, 0f }
			},
			new float[,] { // low roll forward
				{ 0.0f, 0.5f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.6f, 0.4f, 1f, 0f }
			},
			new float[,] { // mid roll forward
				{ 0.25f, 0.75f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.4f, 0.6f, 1f, 0f }
			},
			new float[,] { // mid roll forward
				{ 0.25f, 0.75f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.6f, 0.4f, 1f, 0f }
			},
			new float[,] { // high roll backward
				{ 0.5f, 1.0f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.4f, 0.6f, 1f, 0f }
			},
			new float[,] { // high roll backward
				{ 0.5f, 1.0f, 0f, 0.25f },
				{ 0.5f, 0.5f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.6f, 0.4f, 1f, 0f }
			}
		};
		// Long strokes 
		private float[][,] LongStrokes = new float[5][,]
		{
			new float[,] {
				{ 0.1f, 0.9f, 0f, 0.25f },
				{ 0.8f, 0.2f, 1f, 0.5f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0.2f }
			},
			new float[,] {
				{ 0.1f, 0.9f, 0f, 0.25f },
				{ 0.2f, 0.8f, 1f, 0.5f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0.2f }
			},
			new float[,] {
				{ 0.0f, 0.7f, 0f, 0f },
				{ 0.0f, 1.0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 1.0f, 0.3f, 0f, 0.5f }
			},
			new float[,] {
				{ 0.0f, 0.7f, 0f, 0f },
				{ 0.6f, 0.4f, 0f, 0f },
				{ 0.6f, 0.4f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.5f, 1.0f, 0f, 0.25f }
			},
			new float[,] {
				{ 0.0f, 0.8f, 0f, 0f },
				{ 0.4f, 0.6f, 0f, 0f },
				{ 0.4f, 0.6f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.5f, 1.0f, 0f, 0.25f }
			}
		};
		// Grind strokes
		private float[][,] GrindStrokes = new float[6][,]
		{
			new float[,] { // circular grind
				{ 0.0f, 0.0f, 0f, 0f },
				{ 0.7f, 0.3f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.3f, 0.7f, 0f, 0f }
			},
			new float[,] { // vortex grind reversed
				{ 0.0f, 0.0f, 0f, 0f },
				{ 0.7f, 0.3f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.3f, 0.7f, 0f, 0f }
			},
			new float[,] { // forwad backward grind
				{ 0.0f, 0.1f, 0f, 0f },
				{ 0.0f, 1.0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.6f, 0.4f, 0f, 0f }
			},
			new float[,] { // phased forwad backward grind
				{ 0.0f, 0.1f, 0f, 0f },
				{ 0.0f, 1.0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f }
			},
			new float[,] { // forward-back tilt
				{ 0.0f, 0.2f, 0f, 0f },
				{ 0.2f, 0.8f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.0f, 0f, 0f }
			},
			new float[,] { // forward-back tilt
				{ 0.0f, 0.2f, 0f, 0f },
				{ 0.3f, 0.3f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.1f, 0.7f, 0f, 0f }
			}
		};
		// Tease strokes
		private float[][,] TeaseStrokes = new float[8][,]
		{
			new float[,] { //Orbit tease
				{ 0.9f, 0.9f, 0f, 0f },
				{ 0.7f, 0.3f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.1f, 0.9f, 1f, 0f },
				{ 0.1f, 0.9f, 0f, 0f }
			},
			new float[,] { //Orbit tease reversed
				{ 0.9f, 0.9f, 0f, 0f },
				{ 0.7f, 0.3f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.9f, 0.1f, 1f, 0f },
				{ 0.1f, 0.9f, 0f, 0f }
			},
			new float[,] { //Left-right rock
				{ 0.8f, 0.8f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.9f, 0.1f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 1f, 0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f }
			},
			new float[,] { // Down and back
				{ 0.8f, 1.0f, 0f, 0f },
				{ 0.2f, 0.8f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.9f, 0.3f, 1f, 0f }
			},
			new float[,] { // Back swirl
				{ 0.7f, 0.9f, 0f, 0f },
				{ 0.8f, 0.2f, 0f, 0f },
				{ 0.2f, 0.8f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.2f, 0.8f, 1f, 0f },
				{ 0.3f, 1.0f, 0f, 0f }
			},
			new float[,] { // Back swirl reversed
				{ 0.7f, 0.9f, 0f, 0f },
				{ 0.8f, 0.2f, 0f, 0f },
				{ 0.8f, 0.2f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.8f, 0.2f, 1f, 0f },
				{ 0.3f, 1.0f, 0f, 0f }
			},
			new float[,] { // up-down circle
				{ 0.7f, 1.0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.3f, 0.7f, 1f, 0f },
				{ 0.8f, 0.8f, 0f, 0f }
			},
			new float[,] { // up-down circle reversed
				{ 0.7f, 1.0f, 0f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.5f, 0.5f, 0f, 0f },
				{ 0.7f, 0.3f, 1f, 0f },
				{ 0.8f, 0.8f, 0f, 0f }
			}
		};
		
		// Stroke parameters
		private float strokeVelBase = 0.5f;
		private float strokeVelNext;
		private float strokeVelPrev;
		private float strokeVel;
		private float strokeAng = 6.283185f;
		private float strokeCycles = 1;
		private float[,] arrBase = {{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f }};
		private float[,] arrNext = {{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f }};
		private float[,] arrPrev = {{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f }};
		private float[,] arr = {{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f },{ 0.5f, 0.5f, 0f, 0f }};
		private string strokeType;
		
		// Atom Manipulation Variables
		// Male variables
		protected JSONStorableStringChooser maleChooser;
		protected JSONStorableString maleString;
        private Atom maleObject;
		private FreeControllerV3 malePenisTip;
        private FreeControllerV3 malePenisBase;
        private FreeControllerV3 maleHead;
        private Vector3 malePenisTipX;
        private Vector3 malePenisBaseX;
		private Quaternion malePenisRef;
		private float malePenisLength;
        private Vector3 maleHeadX;

		// Female variables
		protected JSONStorableStringChooser femaleChooser;
        protected JSONStorableString femaleString;
        private Atom femaleObject; 
        private FreeControllerV3 femaleHead;
        private FreeControllerV3 femalePelvis;
		private FreeControllerV3 femaleThighLeft;
		private FreeControllerV3 femaleThighRight;
		private FreeControllerV3 femaleChest;
		
		// Robot position tracking variables
        private float setPositionX;
        private float setPositionY;
        private float setPositionZ;
		private float setPositionRX;
        private float setPositionRY;
        private float setPositionRZ;
        private float setSuck;
	
	    private object serialLock = new object();
        private bool serialIsStarted = false;
	
		
		// Serial Stuff
        SerialPort serial;
		
		// Function to initiate plugin
        public override void Init() {
			
			pluginLabelJSON.val = "Random Stroker";
			
			// Left Column of UI - Serial stuff
			
			// Select COM Port
            portNames = System.IO.Ports.SerialPort.GetPortNames();
            List<string> comPorts = new List<string>(portNames);
            portChooser = new JSONStorableStringChooser("COM Port Chooser", comPorts, "None", "Select COM port", PortChooserCallback);
            UIDynamicPopup portChooserPopup = CreatePopup(portChooser, false);
            portChooserPopup.labelWidth = 300f;
            
            // Set Baud Rate
            baudRate = 115200;
            
            // Serial buttons
            startButton = CreateButton("Start Serial", false);
            if (startButton != null) {
                startButton.button.onClick.AddListener(StartButtonCallback);
            }
            stopButton = CreateButton("Stop Serial", false);
            if (stopButton != null) {
                stopButton.button.onClick.AddListener(StopButtonCallback);
            }
            
            // Window to show what's being sent over serial
            serialReport = new JSONStorableString("Serial Output", "");
            serialOutWindow = CreateTextField(serialReport, false);
    
            // Load/Save axis range limits controls
            autoLoadToggle = new JSONStorableBool("Auto load range limits from device", true);
            CreateToggle(autoLoadToggle, false);
            loadButton = CreateButton("'D2' - Load range limits from device", false);
            if (loadButton != null) {
                loadButton.button.onClick.AddListener(LoadButtonCallback);
            }
            saveButton = CreateButton("'$' - Save range limits to device", false);
            if (saveButton != null) {
                saveButton.button.onClick.AddListener(SaveButtonCallback);
            }
        
            
            axesList = new List<TCodeAxis>();
            axesList.Add(new TCodeAxis("L0","Up",false));
            axesList.Add(new TCodeAxis("L1","Forward",false));
            axesList.Add(new TCodeAxis("L2","Left",true));
            axesList.Add(new TCodeAxis("R0","Twist",true));
            axesList.Add(new TCodeAxis("R1","Roll",true));
            axesList.Add(new TCodeAxis("R2","Pitch",false));
            axesList.Add(new TCodeAxis("A1","Suck",false));
            
            // Draw axis range sliders
            for (int n = 0; n < axesList.Count; n++) {
                CreateSlider(axesList[n].rangeSliderA, false);
                CreateSlider(axesList[n].rangeSliderB, false);
            }
            // Draw axis inverter controls
            for (int n = 0; n < axesList.Count; n++) {
                CreateToggle(axesList[n].invert, false);
            }
            // Draw axis override controls
            for (int n = 0; n < axesList.Count; n++) {
                CreateToggle(axesList[n].manualBool, false);
                CreateSlider(axesList[n].manual, false);
            }
            
            // Toggle system to suppress repeating TCode commands
            suppressRepeats = new JSONStorableBool("Suppress repeat commands", true);
            CreateToggle(suppressRepeats, false);
            // Toggle system to suppress repeating TCode commands
            logSerialIn = new JSONStorableBool("Print incoming serial to log", true);
            CreateToggle(logSerialIn, false);
			
			
			// Right Column of UI 
			
			// Make a list of
			Atom currentAtom;
			List<string> targetChoices = new List<string>();
			foreach (string atomUID in SuperController.singleton.GetAtomUIDs()) {
				currentAtom = SuperController.singleton.GetAtomByUid(atomUID);
				if (atomUID != null && currentAtom.type == "Person") {
					targetChoices.Add(atomUID);
				}
			}
			
			// Set male character nodes
			// Use default male character, or else first on list
			maleString = new JSONStorableString("Male Person","Person#2");
            RegisterString(maleString);
            if (SuperController.singleton.GetAtomByUid(maleString.val) == null) maleString.val = targetChoices[0];
			// Set nodes
			MaleChooserCallback(maleString.val);
			// Chooser menu
			maleChooser = new JSONStorableStringChooser("Reference Chooser", targetChoices, maleString.val, "Select Male",MaleChooserCallback);
            UIDynamicPopup maleChooserPopup = CreatePopup(maleChooser, true);
			maleChooserPopup.labelWidth = 300f;

			// Set female character nodes
			// Use default female character, or else first on list
            femaleString = new JSONStorableString("Female Person","Person");
            RegisterString(femaleString);
            if (SuperController.singleton.GetAtomByUid(femaleString.val) == null) femaleString.val = targetChoices[0];
			// Set nodes
			FemaleChooserCallback(femaleString.val);
			// Chooser menu
			femaleChooser = new JSONStorableStringChooser("Target Chooser", targetChoices, femaleString.val, "Select Female", FemaleChooserCallback);
            UIDynamicPopup femaleChooserPopup = CreatePopup(femaleChooser, true);
			femaleChooserPopup.labelWidth = 300f;

			
		    // Window to show target position
            strokerReport = new JSONStorableString("Target Position", "");
            strokerWindow = CreateTextField(strokerReport, true);
			

			
			
			// Sliders to adjust girl position and movement
            intensityLevel = new JSONStorableFloat("Intensity Level", 0.0f, 0.0f, 1.0f, true, true);
			RegisterFloat(intensityLevel);
			CreateSlider(intensityLevel, true);			
            intensityEscalate = new JSONStorableFloat("Escalation rate", 0.1f, -1.0f, 1.0f, true, true);
			RegisterFloat(intensityEscalate);
			CreateSlider(intensityEscalate, true);
			
			// Weighting sliders
            weightTease = new JSONStorableFloat("Weight - Tease", 1f, 0f, 2f, true, true);
			RegisterFloat(weightTease);
			CreateSlider(weightTease, true);
			weightGrind = new JSONStorableFloat("Weight - Grind", 1f, 0f, 2f, true, true);
			RegisterFloat(weightGrind);
			CreateSlider(weightGrind, true);
			weightLong = new JSONStorableFloat("Weight - Long Strokes", 1f, 0f, 2f, true, true);
			RegisterFloat(weightLong);
			CreateSlider(weightLong, true);
			weightShort = new JSONStorableFloat("Weight - Short Strokes", 1f, 0f, 2f, true, true);
			RegisterFloat(weightShort);
			CreateSlider(weightShort, true);
			
			// Sliders to adjust girl position and movement
			randomVariation = new JSONStorableFloat("Random variation", 0.1f, 0f, 0.5f, true, true);
			RegisterFloat(randomVariation);
			CreateSlider(randomVariation, true);
            adjustGirlUp = new JSONStorableFloat("Adjust Girl Position - Up", 0.0f, -0.2f, 0.2f, true, true);
			RegisterFloat(adjustGirlUp);
			CreateSlider(adjustGirlUp, true);
			adjustGirlForward = new JSONStorableFloat("Adjust Girl Position - Forward", 0.0f, -0.2f, 0.2f, true, true);
			RegisterFloat(adjustGirlForward);
			CreateSlider(adjustGirlForward, true);
			scaleTranslation = new JSONStorableFloat("Scale Translation", 1.0f, 0f, 2f, true, true);
			RegisterFloat(scaleTranslation);
			CreateSlider(scaleTranslation, true);
			scaleRotation = new JSONStorableFloat("Scale Rotation", 1.0f, 0f, 2f, true, true);
			RegisterFloat(scaleRotation);
			CreateSlider(scaleRotation, true);
			
			// Set speeds
			strokeVel = strokeVelBase;
			strokeVelNext = strokeVelBase;
			strokeVelPrev = strokeVelBase;
			
		
        }
		
		
		
		// Update is called with each rendered frame by Unity
        void Update() {
			
			// Do the random stroker stuff
			UpdateRandomStroker();
			
			// Send data over serial
			if (serial != null && serial.IsOpen) {
				SendPulse();
			}
			
			// Do the animation stuff
			UpdateAnimation();
			
		}
		
		
		
		
		
		// Function to the random stroker stuff
		public void UpdateRandomStroker() {

			// Advance stroker one step
			float cycleStep = Time.deltaTime*strokeVel;
			strokeCycles -= cycleStep;
			strokeAng = -strokeCycles*6.283185f; 
			//(Angle is negative because angluar progression should be positive, whilst cycles counts down)

			// Test to see if time to change stroke
			if (strokeCycles < 0) {
				//Escalate intensity
				intensityLevel.val += intensityEscalate.val/10;
				intensityLevel.val = Mathf.Clamp(intensityLevel.val,0f,1f);
				// New stroke cycle - speed
				strokeVelBase = UnityEngine.Random.Range(0.2f + 0.4f*intensityLevel.val,0.4f + 0.8f*intensityLevel.val);
				// New stroke cycle - type
				float chanceTease = weightTease.val*(0.3f - 0.2f*intensityLevel.val);
				float chanceGrind = weightGrind.val*(0.1f + 0.2f*intensityLevel.val);
				float chanceLong = weightLong.val*(0.5f - 0.2f*intensityLevel.val);
				float chanceShort = weightShort.val*(0.1f + 0.2f*intensityLevel.val);
				float choicevalue = UnityEngine.Random.Range(0,chanceTease+chanceGrind+chanceLong+chanceShort);
				int choice = 0;
				if (choicevalue < chanceShort + chanceLong + chanceGrind) { choice = 1; }
				if (choicevalue < chanceShort + chanceLong) { choice = 2; }
				if (choicevalue < chanceShort) { choice = 3; }
				// New stroke cycle - set parameters
				int patternNumber;
				if (choice == 3) {
					patternNumber = UnityEngine.Random.Range(0,ShortStrokes.Length);
					arrBase = ShortStrokes[patternNumber];
					strokeCycles += UnityEngine.Random.Range(4,20);
					strokeVelBase *= 2f;
					strokeType = "Short Strokes #" + (patternNumber+1).ToString();
				} else if (choice == 2) {
					patternNumber = UnityEngine.Random.Range(0,LongStrokes.Length);
					arrBase = LongStrokes[patternNumber];
					strokeCycles += UnityEngine.Random.Range(4,10);
					strokeType = "Long Strokes #" + (patternNumber+1).ToString();
				} else if (choice == 1) {
					patternNumber = UnityEngine.Random.Range(0,GrindStrokes.Length);
					arrBase = GrindStrokes[patternNumber];
					strokeCycles += UnityEngine.Random.Range(3,6);
					strokeVelBase *= 0.75f;
					strokeType = "Grind #" + (patternNumber+1).ToString();
				} else { 
					patternNumber = UnityEngine.Random.Range(0,TeaseStrokes.Length);
					arrBase = TeaseStrokes[patternNumber];
					strokeCycles += UnityEngine.Random.Range(2,8);
					strokeVelBase *= 0.5f;
					strokeType = "Tease #" + (patternNumber+1).ToString();
				}
			}
			
			// Test to see if a new cycle is in progress
			if ( Mathf.FloorToInt(strokeCycles) != Mathf.FloorToInt(strokeCycles+cycleStep)) {
				// Vary speed by random amount
				strokeVelPrev = strokeVelNext;
				strokeVelNext = strokeVelBase*(1+UnityEngine.Random.Range(-2*randomVariation.val,2*randomVariation.val));
				// Install base values and vary by random amount for each stroke
				for (int i = 0; i < 6; i++) {
					float range = Mathf.Abs(arrBase[i,0]-arrBase[i,1]);
					arrPrev[i,0] = arrNext[i,0];
					arrNext[i,0] = arrBase[i,0] + UnityEngine.Random.Range(-range*randomVariation.val,range*randomVariation.val);
					arrNext[i,0] = Mathf.Clamp(arrNext[i,0],0f,1f);
					arrPrev[i,1] = arrNext[i,1];
					arrNext[i,1] = arrBase[i,1] + UnityEngine.Random.Range(-range*randomVariation.val,range*randomVariation.val);
					arrNext[i,1] = Mathf.Clamp(arrNext[i,1],0f,1f);
					arrPrev[i,2] = arrNext[i,2];
					arrNext[i,2] = arrBase[i,2];
					arrPrev[i,3] = arrNext[i,3];
					arrNext[i,3] = arrBase[i,3];
				}
				
			}
			
			float cyclePosition = 1-(strokeCycles - Mathf.Floor(strokeCycles));
			for (int i = 0; i < 6; i++) {
				for (int j = 0; j < 4; j++) {
					arr[i,j] = cyclePosition*arrNext[i,j] + (1-cyclePosition)*arrPrev[i,j];
				}
			}
			
			// Calculate modified speed
			strokeVel = cyclePosition*strokeVelNext + (1-cyclePosition)*strokeVelPrev;

			// Calculate channel output values
			setPositionX = 0.5f*(arr[0,1]+arr[0,0]) - 0.5f*(arr[0,1]-arr[0,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[0,2] + arr[0,3]*Mathf.Sin(strokeAng));
			setPositionY = 0.5f*(arr[1,1]+arr[1,0]) - 0.5f*(arr[1,1]-arr[1,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[1,2] + arr[1,3]*Mathf.Sin(strokeAng));
			setPositionZ = 0.5f*(arr[2,1]+arr[2,0]) - 0.5f*(arr[2,1]-arr[2,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[2,2] + arr[2,3]*Mathf.Sin(strokeAng));
			setPositionRX = 0.5f*(arr[3,1]+arr[3,0]) - 0.5f*(arr[3,1]-arr[3,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[3,2] +arr[3,3]*Mathf.Sin(strokeAng));
			setPositionRY = 0.5f*(arr[4,1]+arr[4,0]) - 0.5f*(arr[4,1]-arr[4,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[4,2] +arr[4,3]*Mathf.Sin(strokeAng));
			setPositionRZ = 0.5f*(arr[5,1]+arr[5,0]) - 0.5f*(arr[5,1]-arr[5,0])*Mathf.Cos(strokeAng + 0.5f*3.14159f*arr[5,2] +arr[5,3]*Mathf.Sin(strokeAng));

            //Send position data to axes
            axesList.First(x => x.channel == "L0").position = setPositionX;
            axesList.First(x => x.channel == "L1").position = setPositionY;
            axesList.First(x => x.channel == "L2").position = setPositionZ;
			axesList.First(x => x.channel == "R0").position = setPositionRX;
            axesList.First(x => x.channel == "R1").position = setPositionRY;
            axesList.First(x => x.channel == "R2").position = setPositionRZ;
            axesList.First(x => x.channel == "A1").position = setSuck;	
			
			// Generate the stroker report window text
			var strokerReportBuilder = new StringBuilder();
			strokerReportBuilder.AppendLine("Stroke pattern: " + strokeType);
			strokerReportBuilder.AppendLine("Strokes remaining: " + strokeCycles.ToString());
			strokerReportBuilder.AppendLine("Stroke cycle: " + cyclePosition.ToString());
			
			strokerReport.val = strokerReportBuilder.ToString();

		}	
		
		// Function to handle the character animation
		public void UpdateAnimation() {

			// Penis (=stroke) reference frame
			malePenisRef = Quaternion.LookRotation((malePenisTip.transform.position - malePenisBase.transform.position),(maleHead.transform.position - malePenisBase.transform.position));
			malePenisLength = (malePenisTip.transform.position - malePenisBase.transform.position).magnitude;

			// Female pelvis angle
			femalePelvis.transform.rotation = Quaternion.LookRotation(malePenisRef* new Vector3(scaleRotation.val*(setPositionRX-0.5f),0.5f,scaleRotation.val*(setPositionRZ-0.5f)),malePenisRef* new Vector3(scaleRotation.val*(setPositionRY-0.5f),0f,1f));
	 
			// Animate the pelvis and thighs
			setPositionZ = (setPositionZ-0.5f)*(0.5f + 0.5f*setPositionX) + 0.5f;
			setPositionY = (setPositionY-0.5f)*(0.5f + 0.5f*setPositionX) + 0.5f;
			femalePelvis.transform.position = malePenisTip.transform.position + malePenisRef* new Vector3(1f*malePenisLength*scaleTranslation.val*(0.5f-setPositionZ),1f*malePenisLength*scaleTranslation.val*(0.5f-setPositionY)+adjustGirlForward.val,malePenisLength*1.5f*(scaleTranslation.val*setPositionX-1f)+adjustGirlUp.val) + femalePelvis.transform.rotation* new Vector3(0f,0.15f,0f);//-0.05f);
			femaleThighLeft.transform.position = femalePelvis.transform.position + femalePelvis.transform.rotation* new Vector3(-0.08f,-0.05f - 0.05f*scaleTranslation.val*setPositionX,0f);
			femaleThighRight.transform.position = femalePelvis.transform.position + femalePelvis.transform.rotation* new Vector3(0.08f,-0.05f - 0.05f*scaleTranslation.val*setPositionX,0f);
			
			// Don't need this yet!
            //femaleHead.transform.position = malePenisBase.transform.position + 2f*malePenisRefZ + new Vector3(0f,0.6f,0f);
		}
		
		
		
		// Function to interpret incoming serial data
        public void InterpretInput(string inString) {
            // 
            switch(inString.Substring(0,1)) {
              case "L":
              case "R":
              case "V":
              case "A":
                string channel = inString.Substring(0,2);
                for (int n = 0; n < axesList.Count; n++) {
                    if (channel == axesList[n].channel) {
                        axesList[n].SetLimitsFromString(inString);
                    }
                }
                break;
              default:
                
                break;
            }
            
            if (inString == "Ready!" && autoLoadToggle.val) {
                SuperController.LogMessage("[Out]: D2");
                lock (serialLock) {
                    serial.WriteLine("D2 \n");
                }
            }
            
        }
		
		// Button callbacks
		protected void PortChooserCallback(string s) {
            portSelected = s;
        }
		protected void StartButtonCallback() {
			StartSerial();
			SuperController.LogMessage("Serial connection started: " + portSelected + ", baud " + baudRate.ToString() );
			if (autoLoadToggle.val) {
                //SuperController.LogMessage("[Out]: D2");
                lock (serialLock) {
                    serial.WriteLine("D2 \n");
                }
            }
        }
		protected void StopButtonCallback() {
			StopSerial();
			SuperController.LogMessage("Serial connection stopped");
		}
		protected void SaveButtonCallback() {
            string sendString;
            if(serial != null && serial.IsOpen) {
                SuperController.LogMessage("Saving axis ranges to T-Code Device:");
                for (int n = 0; n < axesList.Count; n++) {
                    sendString = axesList[n].GetTCodeSaveLimits();
                    SuperController.LogMessage("[Out]: " + sendString);
                    lock (serialLock) {
                        serial.WriteLine(sendString + "\n");
                    }
                }
            }
        }
        protected void LoadButtonCallback() {
            string sendString;
            if(serial != null && serial.IsOpen) {
                SuperController.LogMessage("Identify available T-Code axes:");
                SuperController.LogMessage("[Out]: D2");
                lock (serialLock) {
                    serial.WriteLine("D2 \n");
                }
            }
        }
		protected void MaleChooserCallback(string s) {
			// Set new male object and nodes
			maleString.val = s;
			maleObject = SuperController.singleton.GetAtomByUid(maleString.val);
            malePenisTip = maleObject.GetStorableByID("penisTipControl") as FreeControllerV3;
			malePenisTip.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisTip.currentRotationState = FreeControllerV3.RotationState.On;
            malePenisBase = maleObject.GetStorableByID("penisBaseControl") as FreeControllerV3;
			malePenisBase.currentPositionState = FreeControllerV3.PositionState.On;
			malePenisBase.currentRotationState = FreeControllerV3.RotationState.On;
			maleHead = maleObject.GetStorableByID("headControl") as FreeControllerV3;
        }
 
        protected void FemaleChooserCallback(string s) {
			// Set new female object and nodes
			femaleString.val = s;
			femaleObject = SuperController.singleton.GetAtomByUid(femaleString.val);
            femalePelvis = femaleObject.GetStorableByID("pelvisControl") as FreeControllerV3;
			femalePelvis.currentPositionState = FreeControllerV3.PositionState.On;
			femalePelvis.currentRotationState = FreeControllerV3.RotationState.On;
			femaleThighLeft = femaleObject.GetStorableByID("lThighControl") as FreeControllerV3;
			femaleThighLeft.currentPositionState = FreeControllerV3.PositionState.On;
			femaleThighRight = femaleObject.GetStorableByID("rThighControl") as FreeControllerV3;
			femaleThighRight.currentPositionState = FreeControllerV3.PositionState.On;
            femaleHead = femaleObject.GetStorableByID("headControl") as FreeControllerV3;
			femaleHead.currentPositionState = FreeControllerV3.PositionState.On;
			femaleHead.currentRotationState = FreeControllerV3.RotationState.On;
			femaleChest = femaleObject.GetStorableByID("chestControl") as FreeControllerV3;
			femaleChest.currentRotationState = FreeControllerV3.RotationState.On;
        }
		
		
        // Serial functions
        // Function to start serial
        void StartSerial() {
            if (portSelected != "None") {
                // Add extra characters for COM ports > 9
                if (portSelected.Substring(0,3) == "COM") {
                    if (portSelected.Length != 4) {
                        portSelected = "\\\\.\\" + portSelected;
                    }
                }
                
                // Open the serial connection
                serial = new SerialPort(portSelected, baudRate);
                serial.Open();
                serial.DtrEnable = true;
                serial.ReadTimeout = 10;
                serialIsStarted = true;

				// Start serial thread
                ThreadPool.QueueUserWorkItem(delegate {
                    while (serialIsStarted) {
                        Thread.Sleep(100);
                        if (serial == null || !serial.IsOpen) {
                            return;
                        }

                        // Read incoming serial one line at a time
                        string inString = null;
                        try {
                            lock (serialLock) {
                                inString = serial.ReadLine();
                            }
                        } catch (TimeoutException) {
                            continue;
                        }

                        // Print it to the log if this is active
                        if (logSerialIn.val && inString != null) { SuperController.LogMessage("[In]: " + inString); }
                        // Send each line for interpretation
                        InterpretInput(inString);
                    }
                });
            }
        }
        // Function to stop serial
        void StopSerial() {
            serialIsStarted = false;
            if(serial != null && serial.IsOpen) {
                lock (serialLock) {
                    serial.Close();
                }
            }
        }
        // Function to send device position updates by serial
        void SendPulse() {
            var sendStringBuilder = new StringBuilder();
            var serialReportBuilder = new StringBuilder();
            serialReportBuilder.AppendLine("T-code:");
            // Send one channel sequentially regardless of no change
            sendThisChannelAnyway++;
            if (sendThisChannelAnyway >= axesList.Count) { sendThisChannelAnyway = 0; }
            // For each channel
            for (int n = 0; n < axesList.Count; n++) {
                // Send anyway if it's this channel's turn
                bool suppress = (sendThisChannelAnyway != n);
                // If suppress not set, send anyway
                if (!suppressRepeats.val) { suppress = false;}
                // Acquire T-Code command for this channel
                sendStringBuilder.Append(axesList[n].GetTCode(axesList[n].position,0,suppress));
                serialReportBuilder.Append(axesList[n].lastTCode);
            }

            serialReportBuilder.AppendLine("\n[Out]:");
            serialReportBuilder.Append(sendStringBuilder);
            sendStringBuilder.AppendLine("A90");
            
            serialReport.val = serialReportBuilder.ToString();
            var message = sendStringBuilder.ToString();

            ThreadPool.QueueUserWorkItem(delegate {
                lock (serialLock) {
                    serial.WriteLine(message);
                }
            });
        }
		
		
    // TCode Axis Class
    public class TCodeAxis {
        // Properties
        public string channel;
        public string name;
        public bool symmetric;
        
        // Output parameters
        public float position;
        public string lastTCode;
        
        // Interface parameters
        public JSONStorableFloat rangeSliderA;
        public JSONStorableFloat rangeSliderB;
        public JSONStorableFloat adjust;
        public JSONStorableBool invert;
        public JSONStorableBool manualBool;
        public JSONStorableFloat manual;
        
        // Setup function
        public TCodeAxis(string setChannel, string setName, bool setSymmetric) {
            channel = setChannel.Substring(0,2);
            name = setName;
            symmetric = setSymmetric;
            
            // Set position to default
            position = 0.5f;
            if (channel.Substring(0,1) == "V") { position = 0f; };
            
            // Format output range sliders
            if (!symmetric) {
                rangeSliderA = new JSONStorableFloat("'" + channel + "' (" + name + ") Max", 1f, 0.5f, 1f, true, true);
                rangeSliderB = new JSONStorableFloat("'" + channel + "' (" + name + ") Min", 0f, 0.0f, 0.5f, true, true);
            } else {
                rangeSliderA = new JSONStorableFloat("'" + channel + "' (" + name + ") Max (+/-)", 0.5f, 0f, 0.5f, true, true);
                rangeSliderB = new JSONStorableFloat("'" + channel + "' (" + name + ") Centre", 0.50f, 0.0f, 1f, true, true);
            }
            
            // Format axis inverter
            invert = new JSONStorableBool("Invert '" + channel + "' (" + name + ")", false);
            
            // Format override controls
            manualBool = new JSONStorableBool("Override '" + channel + "' (" + name + ")", false);
            manual = new JSONStorableFloat("'" + channel + "' (" + name + ") Set", 0.5f, 0f, 1f, true, true);
            
            
        }
        
        // Function to create a T-Code string for this axis
        public string GetTCode(float input, int ramp, bool suppressRepeats) {
            float inputLimited, inputScaled;
            // If override, use the override value
            if (manualBool.val) {
                inputScaled = manual.val;
            } else {
                // Use input value; limit input to 0-1 range
                if (input >= 1f) { inputLimited = 1f; }
                else if (input <= 0f) { inputLimited = 0f; }
                else { inputLimited = input; }
                // Scale input according to range limits
                if (!symmetric) {
                    inputScaled = rangeSliderB.val + inputLimited*(rangeSliderA.val-rangeSliderB.val);
                } else {
                    inputScaled = rangeSliderB.val + (inputLimited-0.5f)*2*rangeSliderA.val;
                }
            }
            inputScaled *= 1000;
            string output;
            if (inputScaled >= 999f) {
                output = channel + "999";
            } else if (inputScaled >= 1f) {
                output = channel + inputScaled.ToString("000");
            } else {
                output = channel + "000";
            }
            
            // Live command?
            if (ramp == 0) {
                output += " ";
                // Option to suppress a live T-Code command if it's the same as the last one sent
                if (suppressRepeats && (output == lastTCode)) {
                    output = "";
                } else {
                    lastTCode = output;
                }
            } else {
                // Add ramp to output
                output += "S" + ramp.ToString() + " ";
                lastTCode = output;
            }
                
            return output;
        }
        
        // Function to create a '$' T-Code string
        // This stores the limit preferences to the T-Code device
        public string GetTCodeSaveLimits() {
            string output;
            float min,max;
            if (!symmetric) {
                max = 10000*rangeSliderA.val;
                min = 10000*rangeSliderB.val;
            } else {
                max = 10000*(rangeSliderB.val+rangeSliderA.val);
                min = 10000*(rangeSliderB.val-rangeSliderA.val);
            }
            // '$' to specify range, and identify channel
            output = "$" + channel + "-";
            // Minimum value
            if (min >= 9999f) {
                output += "9999-";
            } else if (min >= 1f) {
                output += min.ToString("0000") + "-";
            } else {
                output += "0000-";
            }
            // Maximum value
            if (max >= 9999f) {
                output += "9999";
            } else if (max >= 1f) {
                output += max.ToString("0000");
            } else {
                output += "0000";
            }
                
            return output;
        }

        // Function to set this axis' limits from incoming serial
        public void SetLimitsFromString(string inString) {
            bool valid = true;
            int min = 0;
            float minF = 0f;
            int max = 1000;
            float maxF = 1f;
            int marker = inString.IndexOf(" ");
            if (marker < 0) { valid = false; }
            if (valid) {
                inString = inString.Substring(marker+1);
                marker = inString.IndexOf(" ");
                if (marker < 0) { valid = false; }
            }
            if (valid) {
                min = int.Parse(inString.Substring(0,marker));
                inString = inString.Substring(marker+1);
                marker = inString.IndexOf(" ");
                if (marker < 0) { valid = false; }
            }
            if (valid) {
                max = int.Parse(inString.Substring(0,marker));
                minF = 0.0001f*(float)min;
                if (minF >= 1f) {minF = 1f;} else if (minF <= 0f) {minF = 0f;}
                maxF = 0.0001f*(float)max;
                if (maxF >= 1f) {maxF = 1f;} else if (maxF <= 0f) {maxF = 0f;}
                if (!symmetric) {
                    rangeSliderA.val = maxF;
                    rangeSliderB.val = minF;
                } else {
                    rangeSliderA.val = 0.5f*(maxF-minF);
                    rangeSliderB.val = 0.5f*(maxF+minF);
                } 
            }

        }
        
    }
		
		
		// Function to close plugin
        void OnDestroy() {
            StopSerial();
        }

		
    }
}